
if (typeof _run_ === 'undefined') { var _run_; }
function ifundef(x,v) {return x === undefined ? v : x;}

class MyBuffer_basic {
  /****************************************************************************************'******** 
    Description:    Class that serves as a basic pixel buffer
    Made by:        Kristian Stormark
    Version:        1.0.0
    > Can be used to copy data to and from a canvas (but is not directly linked to any canvas)
    > Uses the ImageData interface as a super-structure for easy publishing to a canvas etc
    > All rgba values are in [0,255]    
  *************************************************************************************************/
  
  constructor(width,height){    
    // Creates new buffer    
    this.width      = width;                                                                                  // Buffer width
    this.height     = height;                                                                                 // Buffer height
    this.idata      = new ImageData(this.width, this.height);                                                 // Super-structure for easy export
    this.data       = this.idata.data;                                                                        // Main buffer data (Uint8ClampedArray)
  }
  
  clear(rgbx){ 
    // Clears buffer to given rgbx value (or black if not given)
    var i,rgba,k,buffer;
    rgbx            = ifundef(rgbx,[0,0,0]);                                                                  // Default to black
    rgba            = [rgbx[0],rgbx[1],rgbx[2], (rgbx.length>3) ? rgbx[3] : 255];                             // Default to fully opaque    
    for ( k         = 0; k  < this.width*this.height; k++) {                                                  // Loop over buffer (set to rgba)
      i             = k*4; 
      this.data[i+0]= rgba[0];
      this.data[i+1]= rgba[1];
      this.data[i+2]= rgba[2];
      this.data[i+3]= rgba[3];    
    }
  }
  
  putpixel(x,y,rgbx){
    // Sets a pixel in the buffer to the given colour (input rgbx is either rgb or rgba)
    var i,outside;
    x               = Math.round(x);
    y               = Math.round(y);
    outside         = (x<0) || (x>=this.width) || (y<0) || (y>=this.height); if (outside){return;}
    i               = 4 * (x + y * this.width);
    this.data[i+0]  = rgbx[0];
    this.data[i+1]  = rgbx[1];
    this.data[i+2]  = rgbx[2];
    this.data[i+3]  = (rgbx.length>3) ? rgbx[3] : 255;
  }
    
  getpixel(x,y){
    // Gets a pixel in the buffer (returns undefined if outside)
    var i,skip,rgba;
    x               = Math.round(x);
    y               = Math.round(y);
    skip            = (x<0) || (x>=this.width) || (y<0) || (y>=this.height); if (skip){return undefined;}     
    i               = (x + y * this.width) * 4;
    rgba            = [this.data[i+0],this.data[i+1],this.data[i+2],this.data[i+3]];
    return rgba; 
  }
  
}

class MyBuffer extends MyBuffer_basic {  
  /****************************************************************************************'******** 
    Description:    Class that serves as a pixel buffer (extending the basic buffer)
    Made by:        Kristian Stormark
    Version:        1.0.0   
  *************************************************************************************************/
  
  constructor(width,height){        
    _run_           = super(width,height);    
  }
  
  

}

class MyCanvas {  
  /****************************************************************************************'********  
    Description:    Class to work with canvas display - basically a screen to publish buffers onto
    Made by:        Kristian Stormark
    Version:        1.0.0
  *************************************************************************************************/
    
  constructor(canvas){
    // Input is either canvas element or string with element id
    this.el         = typeof canvas === 'string' ? document.getElementById(canvas) : canvas;    
    this.rc2d       = this.el.getContext("2d");
    this.width      = this.el.width;
    this.height     = this.el.height;
  }

  publish(source){
    // Publishes buffer (ImageData / MyBuffer.idata)
    if (source instanceof ImageData)      {this.rc2d.putImageData(source,       0, 0 ); return;}
    if (source instanceof MyBuffer_basic) {this.rc2d.putImageData(source.idata, 0, 0 ); return;}
    _run_           = console.log('Unknown data type');
  }   
  
  resize(width,height){
    // Resizes the canvas
    this.el.width   = width;
    this.el.height  = height;
    this.width      = this.el.width;
    this.height     = this.el.height;
  }
  
  
}

class MyStopwatch {  
  /****************************************************************************************'********
    Description:    Class to work simple timing functions
    Made by:        Kristian Stormark
    Version:        1.0.0
  *************************************************************************************************/  
  
  constructor(){
    this.id         = Math.floor(Math.random() * 1e6);  
    this.t0         = {};    
  }
  
  now(){
    return Date.now();     // Faster than 'new Date().getTime();'
  }
  
  timedata(tag){
    // Returns start-time, current time and duration (if available, 0 otherwise)
    var start,now,dt;
    start           = this.t0[tag];
    now             = this.now();
    dt              = start === undefined ? 0 : now - start;
    return {start,now,dt};
  }
  
  tic(tag){   
    // Sets and returns new start-time (t0) for given tag
    this.t0[tag]    = this.now();
    return this.t0[tag];
  }
  
  toc(tag){
    // Returns elapsed time since start-time for given tag
    return this.timedata(tag).dt;
  }
  
  tac(tag){     
  // Sets new start-time and returns duration (if available, 0 otherwise)
    var data        = this.timedata(tag);    
    this.t0[tag]    = data.now;
    return data.dt;  
  }
  
  adjust(dt){
    // Shifts all timer by dt (typically to adjust for a pause)    
    var list,i;    
    list            = Object.getOwnPropertyNames(this.t0);
    for (i = 0; i< list.length; i++){ this.t0[list[i]] += dt; }
    
  }
  
}

class MyThread {
  /******************************************************************************  
    Description:    Class for running a thread
    Made by:        Kristian Stormark
    Version:        1.0.0
  *****************************************************************************/
  
  constructor(fun,fps){
    this.id         = null;
    this.fun        = fun;
    this.fps        = ifundef(fps,60);    
    this.meter      = {nxtsec:Date.now()+1000, fps:0, cnt:0};
  }
  
  register(){    
    var meter       = this.meter;    
    if (Date.now() > meter.nxtsec){
      meter.nxtsec  = meter.nxtsec + 1000;
      meter.fps     = meter.cnt;
      meter.cnt     = 0;
    }      
    meter.cnt++;    
  }
  
  init(){
    // Intit    
    var sleepint    = Math.round(1000 / this.fps);    
    this.id         = setInterval((function()   {this.fun(); this.register(); }).bind(this), sleepint);  
    return this;
  }
  
  isrunning(){
    // Is the thread running?
    return this.id !==null;
  }
  
  kill(){
    // Kill the thread (if exists)
    if (!this.isrunning()){ return this;}                   // Return if already not running        
    _run_           = clearInterval(this.id); 
    this.id         = null;    
    return this;
  }
  
  start(){
    // Start (if not alrady running)
    return !this.isrunning() ? this.init() : this;          //     
  }
  
  restart(){
    // Restart
    return this.kill().start();
  }
  
}

class MyApplet { 
  /*************************************************************************************************
    Description:    Class for running an "applet"
    Made by:        Kristian Stormark
    Version:        1.0.0
    Comments:       The core functionality is to call the init() function initially (once), and then 
                    for each new frame call the function update() before publishing the buffer onto 
                    the canvas (repeatedly). The init() and update() methods are only abstractly 
                    defined, in the sense that they are empty in the class definition. They will 
                    still be called, but for something to actually happen, they need to be redefined 
                    by any sub-classes or object instances. The property bgbuff is not used by 
                    default, but is included for applets that needs a background buffer (and can be 
                    copied to the main buffer using the bg2buff() function).
  **************************************************************************************************/
  
  constructor(canvas){
    this.canvas     = new MyCanvas(canvas);
    this.buffer     = new MyBuffer(this.canvas.width, this.canvas.height);
    this.bgbuff     = new MyBuffer(this.canvas.width, this.canvas.height);      // Background buffer; not used by default, but available to clients... :)    
    this.thread     = new MyThread( this.nextframe.bind(this) );    
    this.isinit     = false;
    this.stopwatch  = new MyStopwatch();
    this.stoptime   = undefined;
    this.param      = {};                                                       //    
    this.tag        = {on:true, color:'#5C5C5C'}; this.showtag('Not started');  //    
    
  }
    
  isrunning(){
    // Is applet thread running?
    return this.thread.isrunning();
  }  
  
  start(){
    // Start applet   
    if (this.isrunning()) { return this;  }                       // Return if already running        
    if (!this.isinit)     { this.init();  this.isinit = true;}    // Init if not initialised
    if (this.stoptime !== undefined){this.stopwatch.adjust(this.stopwatch.now() - this.stoptime);}
    _run_           = this.thread.init();    
    return this;
  }
  
  stop(){
    // Stop applet
    if (!this.isrunning()){ return this;}                   // Return if already not running    
    this.stoptime   = this.stopwatch.now();     
    _run_           = this.thread.kill();
    _run_           = this.showtag('Stopped');    
    return this;
  }
  
  toggle(){
    // Start/stop applet
    if (this.isrunning()) { return this.stop();   }
    else                  { return this.start();  }    
  }
  
  restart(){
    // Restarts applet (via init())
    _run_           = this.stop();    
    this.stoptime   = undefined;
    this.isinit     = false;    
    return this.start();
  }
  
  fps(fpsx){
    // Sets or gets the fps
    var thread      = this.thread;
    if (fpsx===undefined){
      return thread.meter.fps;
    }
    else{
      _run_         = this.stop();    
      thread.fps    = fpsx;      
      _run_         = this.start();
    }
  }  
  
  showtag(str){
    // Puts a string in the upper left corner
    if(!this.tag.on){ return; }
    var tt          = this.canvas.rc2d;    
    tt.textBaseline = "top";
    tt.strokeStyle  = this.tag.color;
    _run_           = tt.strokeText(str,0,0);    
  }
      
  init(){
    // Initialisation - can be updated by child/instance as needed     
  }
 
  update( ){
    // To be implemented in child class
  }
  
  setparam(key,val){
    // Set parameter value - can be updated by child/instance as needed
    return this;    
  }
  
  nextframe(){
    // Master update function which calles child update + does common ops    
    _run_           = this.update();                                            // Call child update
    _run_           = this.canvas.publish(this.buffer);                         // Publish
    _run_           = this.showtag(this.fps());
  }

  bg2buff(){
    for ( var i     = 0; i<this.buffer.data.length; i++){ this.buffer.data[i] = this.bgbuff.data[i];}
  }

}
